#!/usr/bin/php
<?php

# txt2atascii.php
# Bill Kendrick <bill@newbreedsoftware.com>
# 2014-03-12 - 2014-03-14

# Converts ASCII text, with special escape sequences, into ATASCII text
# for an Atari 8-bit.  Specifically, it's set up to convert an Atari BASIC or
# TurboBASIC XL program into something that can be read in via the 'ENTER'
# command.  As such, it introduces a few extra lines into the output, to:
#  1. 'PRINT' a 'loading...' message
#  2. Obliterate line 10 of the current program -- it is assumed that the
#     program that is being converted to ATASCII will be 'ENTER'ed by a
#     single-line BASIC program
#  3. 'PRINT' a credits message, pause for a moment, and then invoke the
#     'RUN' command

# Escaping features:
#   ` - toggle inverse-video (high-bit of the characters outputted)
#       e.g., "Hello `world`"
#   {#nnn} - A particular byte, declared as 1-3 digits of decimal (0-255)
#   {0nnn} - A particular byte, declared as 1-3 digits of octal (0-377)
#   {$nn} - A particular byte, declared as 1-2 digits of hexadecmal (0-ff)
#   {foo} - Various control and escape characters; consult $escape_sequences
#           (Note, some ignore whether inverse mode (`) is enabled;
#           e.g., {clear} produces CHR$(125), clearscreen; to get
#           CHR$(253), bell, use {bell}.  Others can be made inverse,
#           e.g., {heart} produces CHR$(0), `{heart}` produces CHR$(128).
#   {n foo} - Repeat a control or escape character; e.g. {10 up} produces
#             a sequence of 10 up-arrow characters (CHR$(28))

$escape_sequences = array(
  "heart" => array(0, true),
  "^," => array(0, true),

  "box-vert-right" => array(1, true),
  "^a" => array(1, true),

  "line-right" => array(2, true),
  "^b" => array(2, true),

  "box-up-left" => array(3, true),
  "^c" => array(3, true),

  "box-vert-left" => array(4, true),
  "^d" => array(4, true),

  "box-down-left" => array(5, true),
  "^e" => array(5, true),

  /* FIXME: Named alternative? */
  "/" => array(6, true),
  "^f" => array(6, true),

  /* FIXME: Named alternative? */
  "\\" => array(7, true),
  "^g" => array(7, true),

  "tri-lower-right" => array(8, false),
  "tri-upper-left" => array(8 + 128, false),
  "^h" => array(8, true),

  "block-bottom-right" => array(9, true), /* FIXME: Disallow inverse? */
  /* FIXME: Inverse alternative? */
  "^i" => array(9, true),

  "tri-lower-left" => array(10, false),
  "tri-upper-right" => array(10 + 128, false),
  "^j" => array(10, true),

  "block-top-right" => array(11, true), /* FIXME: Disallow inverse? */
  /* FIXME: Inverse alternative? */
  "^k" => array(11, true),

  "block-top-left" => array(12, true), /* FIXME: Disallow inverse? */
  /* FIXME: Inverse alternative? */
  "^l" => array(12, true),

  "line-top" => array(13, true),
  "^m" => array(13, true),

  "line-bottom" => array(14, true),
  "^n" => array(14, true),

  "block-bottom-left" => array(15, true), /* FIXME: Disallow inverse? */
  /* FIXME: Inverse alternative? */
  "^o" => array(15, true),

  "club" => array(16, true),
  "^p" => array(16, true),

  "box-down-right" => array(17, true),
  "^q" => array(17, true),

  "box-horz" => array(18, true),
  "-" => array(18, true),
  "^r" => array(18, true),

  "box-vert-horz" => array(19, true),
  "^s" => array(19, true),

  "ball" => array(20, true),
  "^t" => array(20, true),

  "block-bottom" => array(21, false),
  "block-top" => array(21 + 128, false),
  "^u" => array(21, true),

  "line-left" => array(22, true),
  "^v" => array(22, true),

  "box-down-horz" => array(23, true),
  "^w" => array(23, true),

  "box-up-horz" => array(24, true),
  "^x" => array(24, true),

  "block-left" => array(25, false),
  "block-right" => array(25 + 128, false),
  "^y" => array(25, true),

  "box-up-right" => array(26, true),
  "^z" => array(26, true),

  "esc" => array(27, false),

  "up" => array(28, false),
  "^-" => array(28, false),
  "del-line" => array(28 + 128, false),
  "sh+bksp" => array(28 + 128, false),

  "down" => array(29, false),
  "^=" => array(29, false),
  "ins-line" => array(29 + 128, false),
  "sh+>" => array(29 + 128, false),

  "left" => array(30, false),
  "^+" => array(30, false),
  "tab-clr" => array(30 + 128, false),
  "^tab" => array(30 + 128, false),

  "right" => array(31, false),
  "^*" => array(31, false),
  "tab-set" => array(31 + 128, false),
  "sh+tab" => array(31 + 128, false),

  "diamond" => array(96, true),
  "^." => array(96, true),

  "spade" => array(123, true),
  "^;" => array(123, true),

  "|" => array(124, true), /* silly, since same as '|', but for consistency */
  "box-vert" => array(124, true),

  "clear" => array(125, false),
  "^<" => array(125, false),
  "sh+<" => array(125, false),
  "bell" => array(125 + 128, false),
  "^2" => array(125 + 128, false),

  "bksp" => array(126, false),
  "del-char" => array(126 + 128, false),
  "^bksp" => array(126 + 128, false),

  "tab" => array(127, false),
  "ins-char" => array(127 + 128, false),
  "^>" => array(127 + 128, false),
);

$fi = STDIN;
$fi_opened = false;
$fo = STDOUT;
$fo_opened = false;

if ($argc == 2 || $argc == 3) {
  if ($argv[1] != '-') {
    $fi = fopen($argv[1], "r");
    if ($fi === false) {
      exit(1);
    }
    $fi_opened = true;
  }

  if ($argc == 3) {
    if ($argv[2] != '-') {
      $fo = fopen($argv[2], "w");
      if ($fo === false) {
        if ($fi_opened) {
          fclose($fi);
        }
        exit(1);
      }
      $fo_opened = true;
    }
  }
} else if ($argc > 3) {
  fprintf(STDERR, "\nUsage: %s [infile [outfile]] {use '-' for stdin/stdout)\n",
    $argv[0]);
  exit(1);
}

fputs($fo, "PRINT \"Converted using txt2atascii\"\x9B");
fputs($fo, "10\x9B");

$lineno = 0;

do {
  $line = rtrim(fgets($fi));

  if (!feof($fi)) {
    $lineno++;

    $inv = false;
    $inbrace = false;
    $esc = "";
    $char_count = 0;

    for ($i = 0; $i < strlen($line); $i++) {
      $c = $line[$i];

      if ($c == '`') {
        /* Toggling inverse-video */

        $inv = !$inv;
      } else if ($c == "{") {
        /* Beginning a braced escape */

        $inbrace = true;
        $esc = "";
      } else if ($inbrace) {
        /* Within a braced escape */

        if ($c == "}") {
          /* Closed the braces; echo the character */

          $inbrace = false;

          $cnt = 1;

          /* Look for repeating sequence value at the beginning */

          if (preg_match("/^([0-9]+) (.*)$/", $esc, $matches) == 1) {
            if ($matches[1] == 0) {
              fprintf(STDERR, "\nError: Escape sequence '%s' has a bad repeat value on line %d:\n%s\n", $esc, $lineno, $line);
              exit(1);
            }

            $cnt = $matches[1];
            $esc = $matches[2];
          }


          /* Figure out what character they want: */

          if (preg_match('/^#([0-9]+)$/', $esc, $matches) == 1) {
            /* Decimal byte */

            $c = intval($matches[1]);
            if ($c > 255) {
              fprintf(STDERR, "\nError: Decimal byte '%s' out of range (0-255)i on line %d:\n%s\n", $esc, $lineno, $line);
              exit(1);
            }
          } else if (preg_match('/^0([0-7]+)$/', $esc, $matches) == 1) {
            /* Octal byte */

            $c = intval(octdec($matches[1]));
            if ($c > 255) {
              fprintf(STDERR, "\nError: Octal byte '%s' out of range (0-255)i on line %d:\n%s\n", $esc, $lineno, $line);
              exit(1);
            }
          } else if (preg_match('/^\$([0-9a-f]+)$/i', $esc, $matches) == 1) {
            /* Hex byte */

            $c = intval(hexdec($matches[1]));
            if ($c > 255) {
              fprintf(STDERR, "\nError: Hexadecimal byte '%s' out of range (0-255)i on line %d:\n%s\n", $esc, $lineno, $line);
              exit(1);
            }
          } else if (array_key_exists($esc, $escape_sequences)) {
            /* Labelled escape or control character */

            list($c, $can_be_inv) = $escape_sequences[$esc];

            if ($inv && $can_be_inv) {
              $c = $c + 128;
            }
          } else {
            /* Don't know! */

            fprintf(STDERR, "\nError: Unknown escape sequence '%s' on line %d:\n%s\n\n",
              $esc, $lineno, $line);
            exit(1);
          }


          for ($n = 0; $n < $cnt; $n++) {
            fputs($fo, chr($c));

            $char_count++;
          }
        } else {
          /* Still within the braces; collect the escape sequence */

          $esc .= $c;
        }
      } else {
        /* Plain characters */

        $c = ord($c);

        if ($inv) {
          $c = $c + 128;
        }

        fputs($fo, chr($c));

        $char_count++;
      }
    }

    fputs($fo, "\x9B");

    if ($char_count > 256) {
      fprintf(STDERR, "\nError: Line %d comes out to more than 256 characters:\n%s\n\n",
              $lineno, $line);
      exit(1);
    } else if ($char_count > 120) {
      fprintf(STDERR, "\nWarning: Line %d comes out to more than 120 characters:\n%s\n\n",
              $lineno, $line);
    }
  }
} while (!feof($fi));

// fputs($fo, "SAVE \"D:AUTORUN.BAS\"\x9B");
fputs($fo, "POKE 180,0:RUN\x9B");

if ($fi_opened) {
  fclose($fi);
}
if ($fo_opened) {
  fclose($fo);
}

